/* $Header: /fridge/cvs/libjcgi/cgi-validate.c,v 1.12 2001/12/10 20:42:49 justins Exp $ */
/*
   
   JCGI - cgi-validate.c      Copyright(c) 2001 Justin David Smith
   justins(at)chaos2.org      http://chaos2.org/
    
   Miscellaneous validation functions
    

   This program is free software; you can redistribute it and/or modify 
   it under the terms of the GNU General Public License as published by 
   the Free Software Foundation, version 2 of the License ONLY.

   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <assert.h>
#include <libjcgi.h>


bool cgi_check_input_chars_html(stringa *input) {
/* Make sure the input is non-NULL, and doesn't contain
   illegal characters.  Any illegal characters found will
   be replaced by instances of "?", and false will be
   returned.  The input will be modified in place.  NOTE
   that this prepares input for emission into an HTML page
   but not necessarily for emission as a URL. */

   const string_pair rewrites[] = 
       { { "<",   "&lt;" },
         { ">",   "&gt;" },
         { "\"",  "&quot;" },
         { "&",   "&amp;" },
         { "|",   "&#124;" },
         { "%",   "&#37;" },
         { NULL,  NULL } };
   char buf[0x1000];
   unsigned char ch;
   bool chars_ok = true;
   int i;

   /* Validate input arguments */
   if(input == NULL) return(false);

   /* Apply the standard rewrites */
   if(strreplacea_list(input, rewrites) > 0) chars_ok = false;
   
   /* Check for special characters in the input */
   i = 0;
   while((ch = (unsigned char)stra(input)[i]) != '\0') {
      if(ch < 0x20 || ch > 0x7E) {
         /* Input was outside normal ASCII range, reject. */
         sbprintf(buf, sizeof(buf), "#%02X", ch);
         strdelinsa(input, i, 1, buf);
         i += strlenn(buf);
         chars_ok = false;
      } else {
         ++i;
      }
   }
   
   return(chars_ok);
   
}


bool cgi_check_input_chars_cgi(stringa *input) {
/* Similar to above, but this prepares the input string
   for emission in a URL by %-encoding suspicious chars. 
   WARNING: input should be processed by the above function
   FIRST. */

   char buf[0x1000];
   unsigned char ch;
   bool chars_ok = true;
   int i;

   /* Validate input arguments */
   if(input == NULL) return(false);
   
   /* Check for special characters in the input */
   i = 0;
   while((ch = (unsigned char)stra(input)[i]) != '\0') {
      /* First few are HTML escapes; % is a guard against
         any yet-unpacked CGI chars; | is special in our
         data/logfiles so we disallow it as well. */
      if((ch >= 'a' && ch <= 'z') ||
         (ch >= 'A' && ch <= 'Z') ||
         (ch >= '0' && ch <= '9') ||
         (ch == '_' || ch == '-') ||
         (ch == '.' || ch == '/') ||
         (ch == ':' || ch == '?') ||
         (ch == '&' || ch == '=') ||
         (ch == ';' || ch == '#') ||
         (ch == '+')) {
         /* Pass through unaltered */
         ++i;
      } else if(ch == ' ') {
         /* Pass spaces through as + */
         stra(input)[i] = '+';
         ++i;
      } else {
         /* Input was outside range of legal chars; rewrite using % */
         sbprintf(buf, sizeof(buf), "%%%02X", ch);
         strdelinsa(input, i, 1, buf);
         chars_ok = false;
         i += strlenn(buf);
      }
   }
   
   return(chars_ok);
   
}


bool cgi_check_file_path(const char *path, bool dir_allowed) {
/* Checks that the input path given contains no offensive
   sequences, e.g. "../", or references to any dot-files.
   The path need not be qualified yet, at this point. */
   
   bool last_was_slash = true;
   
   assert(path != NULL);
   
   while(*path != '\0') {
      if(*path == '.' && last_was_slash) {
         /* We don't permit references to dotfiles */
         return(false);
      } else {
         last_was_slash = false;
         if(*path == '.' && *(path + 1) == '.') {
            /* Consecutive dots are right out */
            return(false);
         } else if(*path == '/') {
            /* Legal here, but may not be legal later */
            if(!dir_allowed) return(false);
            last_was_slash = true;
         } else if((*path < 'a' || *path > 'z') &&
                   (*path < 'A' || *path > 'Z') &&
                   (*path < '0' || *path > '9') &&
                   (*path != '.' && *path != '-') &&
                   (*path != '_')) {
            /* Filename contains illegal chars */
            return(false);
         }
         /* Continue to next character */
         ++path;
      } /* Accessing dotfile or normal file? */
   } /* Scan through pathname */
   
   /* No errors encountered */
   return(true);
   
}
